import logging
import sys
import os


def rital():
    try:
        from datamaestro import prepare_dataset
    except ModuleNotFoundError:
        logging.info("Datamaestro n'est pas installé (cela ne devrait pas arriver)")
        sys.exit(1)

    for dataset_id in [
        "com.github.aagohary.canard",
        "irds.antique.train",
        "irds.antique.test",
    ]:
        logging.info("Preparing %s", dataset_id)
        prepare_dataset(dataset_id)

    try:
        from transformers import AutoTokenizer, AutoModelForMaskedLM
    except ModuleNotFoundError:
        logging.info("Datamaestro n'est pas installé (cela ne devrait pas arriver)")
        sys.exit(1)

    for hf_id in ["Luyu/co-condenser-marco", "huawei-noah/TinyBERT_General_4L_312D"]:
        AutoTokenizer.from_pretrained(hf_id)
        AutoModelForMaskedLM.from_pretrained(hf_id)


def llm():
    try:
        from transformers import (
            AutoTokenizer,
            AutoModelForCausalLM,
            AutoModelForSequenceClassification,
        )
    except ModuleNotFoundError:
        logging.info("Datamaestro n'est pas installé (cela ne devrait pas arriver)")
        sys.exit(1)

    try:
        import datasets
    except ModuleNotFoundError:
        logging.info("datasets n'est pas installé (cela ne devrait pas arriver)")
        sys.exit(1)

    HF_MODELS = [
        # Course 2
        ("TinyLlama/TinyLlama-1.1B-Chat-v1.0", AutoModelForCausalLM),
        ("distilbert-base-uncased", AutoModelForSequenceClassification),
        (
            "distilbert-base-uncased-finetuned-sst-2-english",
            AutoModelForSequenceClassification,
        ),
    ]
    for hf_id, base_class in HF_MODELS:
        AutoTokenizer.from_pretrained(hf_id)
        base_class.from_pretrained(hf_id)

    # IMDB
    datasets.load_dataset("imdb", split="train")


def amal():
    try:
        from datamaestro import prepare_dataset
    except ModuleNotFoundError:
        logging.info("Datamaestro n'est pas installé (cela ne devrait pas arriver)")
        sys.exit(1)

    try:
        import datasets
    except ModuleNotFoundError:
        logging.info("datasets n'est pas installé (cela ne devrait pas arriver)")
        sys.exit(1)

    # From datamaestro
    for dataset_id in [
        "com.lecun.mnist",
        "edu.uci.boston",
        "org.universaldependencies.french.gsd",
        "edu.stanford.aclimdb",
        "edu.stanford.glove.6b.50",
    ]:
        logging.info("Preparing %s", dataset_id)
        prepare_dataset(dataset_id)

    # Hugging Face (CelebA)
    datasets.load_dataset(
        "nielsr/CelebA-faces", os.environ.get("HF_DATASETS_CACHEDIR", None)
    )



import logging
import threading

import gymnasium as gym
import numpy as np
import plotly.graph_objects as go
from dash import Dash, Input, Output, dcc, html, Patch
from pystk2_gymnasium.utils import rotate
import pystk2


class LivePlotServer:
    def __init__(self, host="127.0.0.1", port=8050, refresh_interval=1):
        self.host = host
        self.port = port
        self._lock = threading.Lock()
        self._fig = go.Figure()
        self._trace_count = 0
        self._trace_updates = {}  # pending updates
        logging.getLogger("werkzeug").setLevel(logging.WARNING)  # or ERROR, CRITICAL

        # Allow subclass to setup the initial figure
        if hasattr(self, "setup_figure") and callable(self.setup_figure):
            self.setup_figure(self._fig)

        # Initialize Dash
        self.app = Dash(__name__)
        self.app.layout = html.Div(
            [
                dcc.Graph(
                    id="live-plot",
                    figure=self._fig,
                    style={"width": "100%", "height": "90vh"},
                ),
                dcc.Interval(
                    id="interval", interval=int(refresh_interval * 1000), n_intervals=0
                ),
            ]
        )

        # Callback: triggered every interval
        @self.app.callback(
            Output("live-plot", "figure"), Input("interval", "n_intervals")
        )
        def _refresh(n_intervals):
            with self._lock:
                if not self._trace_updates:
                    return Patch()  # nothing to update

                patch = Patch()
                for idx, data in self._trace_updates.items():
                    patch.data[idx].x = data.get("x", [])
                    patch.data[idx].y = data.get("y", [])
                    patch.data[idx].z = data.get("z", [])
                self._trace_updates.clear()
                return patch

    # ------------------------
    # Server control
    # ------------------------
    def start(self, block=False):
        from werkzeug.serving import make_server

        self._server = make_server(self.host, self.port, self.app.server)
        if block:
            self._server.serve_forever()
        else:

            def run():
                self._server.serve_forever()

            self._thread = threading.Thread(target=run, daemon=True)
            self._thread.start()
        return f"http://{self.host}:{self.port}/"

    def close(self):
        if hasattr(self, "_server"):
            self._server.shutdown()
            if hasattr(self, "_thread"):
                self._thread.join()

    # ------------------------
    # Trace management
    # ------------------------
    def add_trace(self, trace):
        """Add a new trace to the figure."""
        with self._lock:
            self._fig.add_trace(trace)
            idx = self._trace_count
            self._trace_count += 1
            return idx

    def set_trace(self, trace_index, x=None, y=None, z=None):
        """Set or replace the full data of a trace (applied on next refresh)."""
        with self._lock:
            self._trace_updates[trace_index] = {
                "x": None if x is None else x,
                "y": None if y is None else y,
                "z": None if z is None else z,
            }


def quaternion_inverse(q):
    """
    Compute the inverse of a quaternion.
    q = [w, x, y, z]
    """
    w, x, y, z = q
    norm_sq = w**2 + x**2 + y**2 + z**2
    return np.array([w, -x, -y, -z]) / norm_sq


class STKLivePlotServer(LivePlotServer):
    def __init__(self, env: gym.Env, **kwargs):
        self.env = env
        self.top_tracks = 5

        self.track_traces_indices: list[int] = []
        self.item_indices: dict[str, int] = {}
        self.kart_trace_index = 0
        self.other_karts_index = 0
        super().__init__(**kwargs)

    def update_plot(self, obs):
        kart_ix = self.env.kart_indices[0]
        kart = self.env.world.karts[kart_ix]
        rotation = quaternion_inverse(kart.rotation)

        def global_coordinates(x):
            """Returns a vector in the global view

            X right, Y up, Z forwards
            """
            return rotate(x, rotation) + kart.location

        starts = np.vstack(
            [global_coordinates(p) for p in obs["paths_start"][: self.top_tracks]]
        )
        ends = np.vstack(
            [global_coordinates(p) for p in obs["paths_end"][: self.top_tracks]]
        )
        for ix, start, end in zip(self.track_traces_indices, starts, ends):
            self.set_trace(
                ix,
                x=[start[0], end[0]],
                y=[start[1], end[1]],
                z=[start[2], end[2]],
            )

        # Add kart marker (yellow)
        kart_pos = global_coordinates(obs["front"])
        self.set_trace(
            self.kart_trace_index,
            x=[kart_pos[0]],
            y=[kart_pos[1]],
            z=[kart_pos[2]],
        )

        # Add other karts
        other_karts_pos = np.vstack(
            [
                global_coordinates(other_kart_pos)
                for other_kart_pos in obs["karts_position"]
            ]
        )
        self.set_trace(
            self.other_karts_index,
            x=other_karts_pos[:, 0],
            y=other_karts_pos[:, 1],
            z=other_karts_pos[:, 2],
        )

        # Set items
        items_position = np.vstack(
            [global_coordinates(p) for p in obs["items_position"]]
        )
        items_type = np.array(obs["items_type"])
        for v in pystk2.Item.Type.__members__.values():
            trace_ix = self.item_indices[v.name]
            pos = items_position[items_type == v.value]
            self.set_trace(trace_ix, x=pos[:, 0], y=pos[:, 1], z=pos[:, 2])

    def setup_figure(self, fig) -> go.Figure:
        """
        Visualize a 3D track with point indices labeled.

        Parameters:
        -----------
        track_points : numpy.ndarray
            Array of shape (N, 3) representing 3D points
        label_every_n : int
            Label every nth point (default: 5)
        show_all_points : bool
            Whether to show all points as markers (default: True)
        show_line : bool
            Whether to show connecting line (default: True)
        title : str
            Plot title
        """

        # numpy array with shape (N, 3)
        track: pystk2.Track = self.env.track
        track_points = track.path_nodes[:, 0]

        # Add the track line
        for ix in range(len(track.path_nodes)):
            self.add_trace(
                go.Scatter3d(
                    x=track.path_nodes[ix, :, 0],
                    y=track.path_nodes[ix, :, 1],
                    z=track.path_nodes[ix, :, 2],
                    mode="lines",
                    line=dict(color="lightblue", width=4),
                    showlegend=False,
                )
            )

        self.add_trace(
            go.Scatter3d(
                x=track_points[:, 0],
                y=track_points[:, 1],
                z=track_points[:, 2],
                mode="markers",
                marker=dict(size=3, color="red", opacity=0.6),
                name="All Points",
                hovertemplate="Point %{pointNumber}<br>"
                + "X: %{x:.3f}<br>"
                + "Y: %{y:.3f}<br>"
                + "Z: %{z:.3f}<extra></extra>",
            )
        )

        # Add start marker (green)
        self.add_trace(
            go.Scatter3d(
                x=[track_points[0, 0]],
                y=[track_points[0, 1]],
                z=[track_points[0, 2]],
                mode="markers+text",
                marker=dict(size=10, color="green", symbol="diamond"),
                text=["START"],
                textposition="top center",
                textfont=dict(size=14, color="green"),
                name="Start Point",
                hovertemplate="Start Point<br>"
                + "X: %{x:.3f}<br>"
                + "Y: %{y:.3f}<br>"
                + "Z: %{z:.3f}<extra></extra>",
            )
        )

        # Paths next to the kart
        for _ in range(self.top_tracks):
            self.track_traces_indices.append(
                self.add_trace(
                    go.Scatter3d(
                        x=[],
                        y=[],
                        z=[],
                        mode="lines",
                        line=dict(color="yellow", width=8),
                        showlegend=False,
                    )
                )
            )

        # The kart itself
        self.kart_trace_index = self.add_trace(
            go.Scatter3d(
                x=[],
                y=[],
                z=[],
                mode="markers+text",
                marker=dict(size=10, color="orange", symbol="diamond"),
                text=["Kart"],
                textposition="top center",
                textfont=dict(size=14, color="orange"),
                name="Kart",
            )
        )

        # The enemy karts
        self.other_karts_index = self.add_trace(
            go.Scatter3d(
                x=[],
                y=[],
                z=[],
                mode="markers",
                marker=dict(size=10, color="red", symbol="x"),
                text=["Other kart"],
                textposition="top center",
                textfont=dict(size=14, color="red"),
                name="Other",
            )
        )

        # Add items
        for v in pystk2.Item.Type.__members__.values():
            self.item_indices[v.name] = self.add_trace(
                go.Scatter3d(
                    x=[],
                    y=[],
                    z=[],
                    mode="text",
                    text=[v.name],
                    textfont=dict(size=12, color="black"),
                )
            )

        # Update layout for better visualization
        fig.update_layout(
            title=dict(text="STK Track", x=0.5, font=dict(size=16)),
            scene=dict(
                aspectmode="data",
                xaxis_title="X Coordinate",
                yaxis_title="Y Coordinate",
                zaxis_title="Z Coordinate",
                zaxis=dict(autorange="reversed"),
                # camera position (data domain => everything is between -1 and 1)
                camera=dict(
                    eye=dict(x=1.5, y=1.5, z=2.0),
                    center=dict(x=0, y=0, z=0),
                    up=dict(x=0, y=1, z=0),
                ),
            ),
            autosize=True,
            # width=1000,
            # height=700,
            showlegend=True,
        )



import logging
import shutil
import subprocess
import sys


def get_pip_command():
    """Detect and return the appropriate pip command (uv or pip)"""
    if shutil.which("uv"):
        return ["uv", "pip", "install"]
    return [sys.executable, "-m", "pip", "install"]


def install_courses(courses):
    """Install su_master_mind with the specified course extras

    Performs all pre-flight checks and then installs all courses in a single command.

    Args:
        courses: List of course names to install (e.g., ['rl', 'deepl'])
    """
    if not courses:
        logging.info("No courses to install")
        return

    # Deduplicate and sort courses
    unique_courses = sorted(set(courses))

    # Perform all pre-flight checks before installing
    # Check for swig if RL course is in the list
    if "rl" in unique_courses:
        if sys.platform == "win32":
            has_swig = shutil.which("swig.exe")
        else:
            has_swig = shutil.which("swig")

        if not has_swig:
            logging.error(
                "swig n'est pas installé: sous linux utilisez le "
                "gestionnaire de paquets:\n - sous windows/conda : "
                "conda install swig\n - sous ubuntu: sudo apt install swig"
            )
            sys.exit(1)

    # Build the package specification with extras
    extras = ",".join(unique_courses)
    package_spec = f"su_master_mind[{extras}]"

    # Get the pip command
    pip_cmd = get_pip_command()

    # Install the package with extras (single command for all courses)
    cmd = pip_cmd + [package_spec]
    logging.info(f"Installing courses {extras}: {' '.join(cmd)}")

    try:
        subprocess.check_call(cmd)
        logging.info(f"Successfully installed {package_spec}")
    except subprocess.CalledProcessError as e:
        logging.error(f"Failed to install {package_spec}: {e}")
        sys.exit(1)



try:
    from functools import cached_property
except ImportError:
    from cached_property import cached_property
import json
import logging
from pathlib import Path
from appdirs import user_config_dir

COURSES = set(["llm", "rl", "rital", "deepl", "adl"])


class Configuration:
    def __init__(self):
        logging.info(f"Reading configuration from {self.path}")
        config = {}
        if self.path.exists():
            with self.path.open("r") as fp:
                config = json.load(fp)

        self.courses = set()
        for course in config.get("courses", []):

            if course in COURSES:
                self.courses.add(course)
            else:
                logging.warning("Course '%s' does not exit – removing", course)

    def save(self):
        self.path.parent.mkdir(exist_ok=True, parents=True)
        s = json.dumps({"courses": [c for c in self.courses]})
        self.path.write_text(s)
        logging.info("Wrote configuration in %s", self.path)

    @cached_property
    def path(self) -> Path:
        return Path(user_config_dir("master-mind", "isir")) / "config.json"



from pathlib import Path
from typing import List
import click
import sys
import importlib
import importlib.metadata
import logging
from .configuration import Configuration
from .utils import check_last_mastermind
from . import install
from packaging.version import parse as parse_version

logging.basicConfig(level=logging.INFO)

original_args = [arg for arg in sys.argv]
CLASS_NAMES = ["deepl", "rl", "llm", "adl", "rital"]


@click.group()
def main():
    pass


@main.group()
def courses():
    """Permet de gérer la liste des cours suivis"""
    pass


@click.argument("courses", type=click.Choice(CLASS_NAMES), nargs=-1)
@courses.command("add")
def courses_add(courses: List[str]):
    """Ajout de cours"""
    configuration = Configuration()
    configuration.courses.update(courses)
    configuration.save()
    print(  # noqa: T201
        "Don't forget to update the python packages with `master-mind update`"
    )


@click.argument("courses", type=click.Choice(CLASS_NAMES), nargs=-1)
@courses.command("rm")
def courses_rm(courses: List[str]):
    """Enlever un cours"""
    configuration = Configuration()
    for course in courses:
        configuration.courses.remove(course)
    configuration.save()


@courses.command("list")
def courses_list():
    """Liste des cours"""
    for course in Configuration().courses:
        print(course)  # noqa: T201


@click.option("--no-self-update", is_flag=True)
@main.command()
def update(no_self_update: bool):
    """Mettre à jour l'ensemble des modules pour les cours suivis"""
    if not no_self_update:
        check_last_mastermind(original_args)

    # Install all configured courses in a single command
    configuration = Configuration()
    install.install_courses(list(configuration.courses))

    print(  # noqa: T201
        "Don't forget to download the datasets with `master-mind download-datasets`"
    )


@main.command()
def download_datasets():
    """Mettre à jour l'ensemble des jeux de données pour les cours suivis"""
    import os

    # Ensure HuggingFace Hub is online for dataset downloads
    hf_hub_offline = os.environ.pop("HF_HUB_OFFLINE", None)

    try:
        configuration = Configuration()
        datasets = importlib.import_module("master_mind.datasets")

        for course in configuration.courses:
            logging.info("Installing resources for %s", course)
            getattr(datasets, course, lambda: None)()
    finally:
        # Restore original HF_HUB_OFFLINE setting
        if hf_hub_offline is not None:
            os.environ["HF_HUB_OFFLINE"] = hf_hub_offline


@main.group()
def rl():
    """Commandes spécifiques au module RL"""
    pass


@click.option("--hide", is_flag=True)
@click.option("--debug", is_flag=True)
@click.option("--error-handling", is_flag=True, help="Catch exceptions within agents")
@click.option(
    "--output",
    default=None,
    type=Path,
    help="Path for output JSON file (none = standard output)",
)
@click.option(
    "--no-check", is_flag=True, help="Do not check if master-mind is up to date"
)
@click.option(
    "--interaction",
    type=click.Choice(["none", "interactive", "map"]),
    help="Race interaction type (when debugging)",
)
@click.option(
    "--num-karts",
    type=int,
    help="Number of karts (must be greater than number of zip files)",
)
@click.option(
    "--action-timeout",
    type=float,
    default=None,
    help="Maximum time in seconds allowed for each agent action (default: no timeout)",
)
@click.argument("file_or_modules", type=str, nargs=-1)
@rl.command("stk-race")
def rld_stk_race(
    no_check,
    hide,
    debug,
    num_karts,
    file_or_modules,
    interaction,
    output,
    error_handling,
    action_timeout,
):
    """Race"""
    if not no_check:
        check_last_mastermind(original_args)

        version = parse_version(importlib.metadata.version("pystk2_gymnasium"))
        assert version >= parse_version("0.7.2") and version < parse_version(
            "0.8.0"
        ), f"Expected pytstk2-gymnasium version 0.7.*, got {version}. Please update."

    if debug:
        logging.getLogger().setLevel(logging.DEBUG)

    from master_mind.rld.stk import race, InteractionMode

    if num_karts < 1 or len(file_or_modules) == 0:
        logging.error("At least one kart")
        sys.exit(1)

    race(
        hide,
        num_karts,
        file_or_modules,
        interaction=InteractionMode[(interaction or "NONE").upper() ],
        output=output,
        error_handling=error_handling,
        action_timeout=action_timeout,
    )
